#!/usr/bin/env ruby

# Copyright (C) 2006-2008 by Sergio Pistone
# sergio_pistone@yahoo.com.ar
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

require File.expand_path( File.dirname( __FILE__ ) + "/../utils/strings" )
require File.expand_path( File.dirname( __FILE__ ) + "/../utils/xmlhash" )
require File.expand_path( File.dirname( __FILE__ ) + "/../utils/logging" )
require File.expand_path( File.dirname( __FILE__ ) + "/../utils/kde" )
require File.expand_path( File.dirname( __FILE__ ) + "/../i18n/i18n" )
require File.expand_path( File.dirname( __FILE__ ) + "/../gui/gui" )
require File.expand_path( File.dirname( __FILE__ ) + "/../lyrics" )
require File.expand_path( File.dirname( __FILE__ ) + "/amarok" )
require File.expand_path( File.dirname( __FILE__ ) + "/plugins" )

require "cgi"
require "thread"

TOOLKIT_PRIORITY = ["qt3", "qt4", "gtk", "tk"]

$LOG_MAX_LINES = 2000
$LOG_FILENAME = "#{ENV["HOME"]}/.wikilyrics.log"

class Application

	@@MENU_ENTRY_SEARCH_CURRENT = I18n.get( "amarok.application.search.current" )
	@@MENU_ENTRY_SEARCH_SELECTED = I18n.get( "amarok.application.search.selected" )
	@@MENU_ENTRY_CLEAR_LYRICS_CACHE = I18n.get( "amarok.application.clearlyricscache" )

	@@SCRIPT_NAME = "Wiki-Lyrics"
	@@CONFIG_FILE = "wikilyrics.xml"

	def initialize()
		self.cleanup_lyrics = true
		self.log_file = nil
	end

	def log_file()
		return @log_file
	end

	def log_file=( filename )
		@log_file = log_file
		Plugins.all_plugins().each() { |plugin| plugin.log_file = log_file() }
	end

	def log?()
		return @log_file != nil
	end

	def log( message, new_lines=1 )
		Logging.log( @log_file, message, new_lines ) if ( @log_file )
	end

	def cleanup_lyrics?()
		return @cleanup_lyrics
	end

	def cleanup_lyrics=( cleanup_lyrics )
		@cleanup_lyrics = cleanup_lyrics
		Plugins.all_plugins().each() { |plugin| plugin.cleanup_lyrics = cleanup_lyrics }
	end

	def notify( message )
		Amarok.notify( "<b>[" + @@SCRIPT_NAME + "]</b> " + message.gsub( "\n", " " ) )
	end

	def popup( message )
		Amarok.popup( "<b>[" + @@SCRIPT_NAME + "]</b><br/>" + message )
	end

	def add_custom_menu_item( menu_item )
		Amarok.add_custom_menu_item( @@SCRIPT_NAME, menu_item )
	end

	def remove_custom_menu_item( menu_item )
		Amarok.remove_custom_menu_item( @@SCRIPT_NAME, menu_item )
	end

	# EVENT HANDLERS

	def read_config()
		values = {
			"cleanup_lyrics" => cleanup_lyrics?(),
			"write_log" => log?(),
			"used_plugins" => Plugins.used_names().join( ";" )
		}
		XMLHash.read( @@CONFIG_FILE, values )
		log = values["write_log"].to_s() == "true"
		new_log_file = log ? $LOG_FILENAME : nil
		Logging.setup( new_log_file, $LOG_MAX_LINES, log_file() )
		self.log_file = new_log_file
		self.cleanup_lyrics = values["cleanup_lyrics"].to_s() == "true"
		Plugins.used_names = values["used_plugins"].split( ";" )
		Plugins.all_plugins().each() { |plugin| plugin.read_config( @@CONFIG_FILE ) }
	end

	def on_start()
		read_config()
		add_custom_menu_item( @@MENU_ENTRY_SEARCH_CURRENT )
		add_custom_menu_item( @@MENU_ENTRY_SEARCH_SELECTED )
		add_custom_menu_item( @@MENU_ENTRY_CLEAR_LYRICS_CACHE )
		Plugins.all_plugins().each() { |plugin| plugin.on_start() }
	end

	def write_config()
		values = {
			"cleanup_lyrics" => cleanup_lyrics?(),
			"write_log" => log?(),
			"used_plugins" => Plugins.used_names().join( ";" )
		}
		XMLHash.write( @@CONFIG_FILE, values )
		Plugins.all_plugins().each() { |plugin| plugin.write_config( @@CONFIG_FILE ) }
	end

	def on_quit()
		write_config()
		remove_custom_menu_item( @@MENU_ENTRY_SEARCH_CURRENT )
		remove_custom_menu_item( @@MENU_ENTRY_SEARCH_SELECTED )
		remove_custom_menu_item( @@MENU_ENTRY_CLEAR_LYRICS_CACHE )
		Plugins.all_plugins().each() { |plugin| plugin.on_quit() }
	end

	def on_configure()

		values = {
			"cleanup_lyrics" => cleanup_lyrics?(),
			"write_log" => log?(),
			"used_plugins" => Plugins.used_names(),
			"script_name" => @@SCRIPT_NAME,
			"unused_plugins" => Plugins.all_names() - Plugins.used_names()
		}
		if ( GUI.show_plugins_manager_dialog( values ) )
			log = values["write_log"].to_s() == "true"
			new_log_file = log ? $LOG_FILENAME : nil
			Logging.setup( new_log_file, $LOG_MAX_LINES, log_file() )
			self.log_file = new_log_file
			self.cleanup_lyrics = values["cleanup_lyrics"].to_s() == "true"
			Plugins.used_names = values["used_plugins"]
		end
		Plugins.all_plugins().each() { |plugin| plugin.configure() }

		write_config()

	end


	def on_search_current_lyrics()
		if ( Amarok.playing?() )
			song_data = Amarok.query_song_data( Amarok.get_current_path() )
			if ( GUI.show_search_lyrics_dialog( song_data ) )
				on_fetch_lyrics( Lyrics::Request.new(
					song_data["artist"],
					song_data["title"],
					song_data["album"],
					song_data["year"]
				) )
			end
		else
			notify( I18n.get( "amarok.application.search.current.nosongplaying" ) )
		end
	end

	def on_search_file_lyrics( file )
		if ( (song_data = Amarok.query_song_data( file )).empty?() )
			notify( I18n.get( "amarok.application.search.noinfofound" ) )
		elsif ( GUI.show_search_lyrics_dialog( song_data ) )
			request = Lyrics::Request.new( song_data["artist"], song_data["title"], song_data["album"], song_data["year"] )
			response, plugin, suggestions = fetch_lyrics( request )
			if ( response.lyrics )
				dialog_data = {
					"artist" =>		response.artist,
					"title" =>		response.title,
					"lyrics" =>		response.lyrics,
					"site_name" =>	plugin.site_name(),
				}
				GUI.show_lyrics_dialog( dialog_data )
			else
				notify( I18n.get( "amarok.application.search.nolyricsfound", song_data["title"], song_data["artist"] ) )
			end
		end
	end

	def on_custom_menu_item_selected( menu, item, urls )
		if ( menu == @@SCRIPT_NAME )
			if ( item == @@MENU_ENTRY_SEARCH_CURRENT )
				on_search_current_lyrics()
			elsif ( item == @@MENU_ENTRY_SEARCH_SELECTED )
				url = URI.parse( urls[0] )
				return if ( url.scheme != "file" )
				on_search_file_lyrics( URI.decode( url.path ) )
			elsif ( item == @@MENU_ENTRY_CLEAR_LYRICS_CACHE )
				if ( GUI.show_confirmation_dialog( I18n.get( "amarok.application.clearlyricscache.confirm" ), @@SCRIPT_NAME) )
					Amarok.query( "DELETE FROM lyrics WHERE url <> ''" )
					notify( I18n.get( "amarok.application.clearlyricscache.done" ) )
				end
			end
		elsif ( plugin = Plugins.plugin_by_name( menu ) )
			plugin.on_custom_menu_item_selected( menu, item, urls )
		end
	end

	def build_lyrics_data_hash( response, plugin=nil )
		return {
			"artist"	=> response.artist ? response.artist : response.request.artist,
			"title"		=> response.title ? response.title : response.request.title,
			"lyrics"	=> response.lyrics,
			"url"		=> response.url,
			"add_url"	=> plugin ? plugin.build_song_add_url( response.request ) : nil,
			"site_name"	=> plugin ? plugin.site_name : nil,
			"site_host"	=> plugin ? plugin.site_host : nil,
		}
	end
	protected :build_lyrics_data_hash

	def wikis_process_response( response, response_plugin, searched_wiki_plugins )
		Plugins.wiki_plugins().each() do |plugin|
			plugin.wiki_process_response( response, response_plugin, searched_wiki_plugins.include?( plugin ) )
		end
	end

	def fetch_lyrics( request )

		used_plugins = Plugins.used_plugins()

		response = nil
		response_plugin = nil
		suggestions = []

		if ( used_plugins.size() == 0 )
			popup( I18n.get( "amarok.application.error.nopluginsselected" ) )
			response = Lyrics::Response.new( request )
		else
			searched_wiki_plugins = []
			used_plugins.each() do |plugin|
				begin
					searched_wiki_plugins << plugin if ( plugin.is_a?( MediaWikiLyrics ) )
					response = plugin.lyrics_direct_search( request )
					if ( response.lyrics )
						response_plugin = plugin
						break
					end
					response, suggs = plugin.lyrics_from_suggestions( request )
					if ( suggs.size > 0 )
						suggestions << plugin.plugin_name()
						suggestions.concat( suggs )
					end
					if ( response.lyrics )
						response_plugin = plugin
						break
					end
				rescue TimeoutError
					notify( I18n.get( "amarok.application.search.plugintimeout", plugin.plugin_name(), plugin.site_host() ) )
				end
			end
			wikis_process_response( response, response_plugin, searched_wiki_plugins )
		end

		return response, response_plugin, suggestions

	end

	def on_fetch_lyrics( request )

		# NOTE: we must check that the current song doesn't change while we're fetching
		# the lyrics, otherwise Amarok will associate the lyrics with the wrong song
		prev_artist = Strings.normalize!( Amarok.get_current_artist().to_s() )
		prev_title  = Strings.normalize!( Amarok.get_current_title().to_s() )

		response, plugin, suggestions = fetch_lyrics( request )
		lyrics_data = build_lyrics_data_hash( response, plugin )

		return if ( Strings.normalize!( Amarok.get_current_artist().to_s() ) != prev_artist ||
					Strings.normalize!( Amarok.get_current_title().to_s() ) != prev_title )

		if ( lyrics_data["lyrics"] )
			Amarok.show_lyrics(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["lyrics"],
				lyrics_data["site_name"],
				lyrics_data["site_host"],
				lyrics_data["url"],
				lyrics_data["add_url"]
			)
		elsif ( suggestions.size > 0 )
			Amarok.show_suggestions(
				lyrics_data["artist"],
				lyrics_data["title"],
				suggestions,
				lyrics_data["add_url"]
			)
		else
			Amarok.show_not_found(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["add_url"]
			)
		end

	end

	def fetch_lyrics_from_url( request, url )

		used_plugins = Plugins.used_plugins()

		response = nil
		response_plugin = nil

		if ( used_plugins.size() == 0 )
			popup( I18n.get( "amarok.application.error.nopluginsselected" ) )
			response = Lyrics::Response.new( request )
		else
			searched_wiki_plugins = []
			used_plugins.each() do |plugin|
				begin
					next if ( ! plugin.known_url?( url ) )
					searched_wiki_plugins << plugin if ( plugin.is_a?( MediaWikiLyrics ) )
					response = plugin.lyrics_from_url( request, url )
					if ( response.lyrics )
						response_plugin = plugin
						break
					end
				rescue TimeoutError
					notify( I18n.get( "amarok.application.search.plugintimeout", plugin.plugin_name(), plugin.site_host() ) )
				end
			end
			wikis_process_response( response, response_plugin, searched_wiki_plugins )
		end

		return response, response_plugin
	end

	def on_fetch_lyrics_from_url( request, url )

		# NOTE: we must check that the current song doesn't change while we're fetching
		# the lyrics, otherwise Amarok will associate the lyrics with the wrong song
		prev_artist = Strings.normalize!( Amarok.get_current_artist().to_s() )
		prev_title  = Strings.normalize!( Amarok.get_current_title().to_s() )

		response, plugin = fetch_lyrics_from_url( request, url )
		lyrics_data = build_lyrics_data_hash( response, plugin )

		return if ( Strings.normalize!( Amarok.get_current_artist().to_s() ) != prev_artist ||
					Strings.normalize!( Amarok.get_current_title().to_s() ) != prev_title )

		if ( lyrics_data["lyrics"] )
			Amarok.show_lyrics(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["lyrics"],
				lyrics_data["site_name"],
				lyrics_data["site_host"],
				lyrics_data["url"],
				lyrics_data["add_url"]
			)
		else
			Amarok.show_not_found(
				lyrics_data["artist"],
				lyrics_data["title"],
				lyrics_data["add_url"]
			)
		end

	end


	def set_configure_msg()
		@mutex.synchronize do
			# prioritize configure messages:
			if ( @messages[0] != "configure" )
				@messages.insert( 0, "configure" )
			end
			@queued_cond.signal()
		end
	end
	protected :set_configure_msg

	def set_generic_msg( command )
		@mutex.synchronize do
			# we queue only two messages to try to stay in synch with Amarok current track
			@messages[@messages.size < 2 ? @messages.size : 1 ] = command
			@queued_cond.signal()
		end
	end
	protected :set_generic_msg

	def get_msg()
		@mutex.synchronize do
			@queued_cond.wait( @mutex ) if ( @messages.empty?() )
			return @messages.shift()
		end
	end
	protected :get_msg

	def exec( threaded_tasks=false )
		@mutex = Mutex.new()
		@queued_cond  = ConditionVariable.new()
		@messages = []

		exit_code = 0

		a = Thread.new() do
			begin
				on_start()
				run_worker( threaded_tasks )
			rescue Errno::ECONNREFUSED => e
				popup( I18n.get( "amarok.application.error.connectionrefused" ) )
				popup( e.backtrace().join( "<br/>" ) )
			rescue SocketError, TimeoutError => e
				popup( I18n.get( "amarok.application.error.connectiondown" ) )
			rescue Exception => e # unexpected error
				$stderr << e.message() << "\n"
				$stderr << e.backtrace().join( "\n" )
				log( e.message() )
				log( e.backtrace().join( "\n" ), 0 )
				exit_code = 1
			end
			exit( exit_code )
		end

		trap( "SIGTERM" ) { $stderr << "SIGTERM caugth" } # causes to fall to the Exception block below on abnormal termination
		begin
			run_stdin_listener()
		rescue Exception => e
		end
		on_quit()
		exit( exit_code )
	end

	def run_stdin_listener()
		loop do
			message = gets().chomp()
			command = /[A-Za-z]*/.match( message ).to_s()
			case command
				when "configure"
					set_configure_msg()
				when "fetchLyrics"
					set_generic_msg( message )
				when "fetchLyricsByUrl"
					set_generic_msg( message )
				when "customMenuClicked"
					set_generic_msg( message )
			end
		end
	end
	protected :run_stdin_listener

	def run_worker( threaded_tasks )
		loop do
			message = get_msg()
			return if ( message == nil )
			command = /[A-Za-z]*/.match( message ).to_s()
			case command
				when "configure"
					on_configure()
				when "fetchLyrics"
					artist = Amarok.get_current_artist()
					title  = Amarok.get_current_title()
					if ( artist && title )
						if ( threaded_tasks )
							Thread.new() do
								on_fetch_lyrics( Lyrics::Request.new(
									Strings.cleanup_artist( artist, title ),
									Strings.cleanup_title( title ),
									Amarok.get_current_album(),
									Amarok.get_current_year()
								) )
							end
						else
							on_fetch_lyrics( Lyrics::Request.new(
								Strings.cleanup_artist( artist, title ),
								Strings.cleanup_title( title ),
								Amarok.get_current_album(),
								Amarok.get_current_year()
							) )
						end
					else
						Amarok.show_error()
					end
				when "fetchLyricsByUrl"
					args = message.split()
					if args.length < 2
						Amarok.show_error()
					else
						artist = Amarok.get_current_artist().to_s()
						title  = Amarok.get_current_title().to_s()
						if ( threaded_tasks )
							Thread.new() do
								on_fetch_lyrics_from_url(
									Lyrics::Request.new(
										Strings.cleanup_artist( artist, title ),
										Strings.cleanup_title( title ),
										Amarok.get_current_album(),
										Amarok.get_current_year()
									),
									args[1]
								)
							end
						else
							on_fetch_lyrics_from_url(
								Lyrics::Request.new(
									Strings.cleanup_artist( artist, title ),
									Strings.cleanup_title( title ),
									Amarok.get_current_album(),
									Amarok.get_current_year()
								),
								args[1]
							)
						end
					end
				when "customMenuClicked"
					urls_start = message.index( /([a-zA-Z]+:\/\/[^: ]+){1,}/ )
					return if ( urls_start == nil )
					menu_id = message.slice( "customMenuClicked:".size()+1..urls_start-2 )
					menu, item = Amarok.get_custom_menu_item( menu_id )
					if ( menu && item )
						urls = message.slice( urls_start..-1 ).split( " " )
						if ( threaded_tasks )
							Thread.new() { on_custom_menu_item_selected( menu, item, urls ) }
						else
							on_custom_menu_item_selected( menu, item, urls )
						end
					end
			end
		end
	end
	protected :run_worker

end


if ( $0 == __FILE__ )

	if ( GUI.set_toolkit_priority( TOOLKIT_PRIORITY ) == nil )
		Amarok.popup( I18n.get( "amarok.application.error.notoolkit" ) )
		exit( 0 )
	end

	proxy_url, excluded_urls, reverse = KDE.get_proxy_settings()
	HTTP.set_proxy_settings( proxy_url, excluded_urls, reverse )

	Application.new().exec()

end
